/*
 *  ps_main.c
 *  ProteinSplitter
 *
 *  Created by Naoto Morikawa on 7/21/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h> /* getopt */
#include "ps_main.h"
#include "ps_pdb_data.h"
#include "ps_print.h"

extern int read_pdb_file(char * filename);
extern void create_files_for_each_segment(char * filename, int seg_length);

struct pe_protein my_protein;
int flat_dir_flag;

/**
 * main - encodes the structure of a protein into a binary sequence.
 * @argc   :
 *
 * SYNOPSIS
 *        ProteinSplitter [-l n] [-d] filename
 *        (n = length of fragments)
 */
int main(int argc, char ** argv)
{
	int c;
	int seg_length;
	char * filename;
	
	seg_length = 5;
	flat_dir_flag = 1;	/* 0 if structured dir, else 1 */
	while ((c = getopt(argc, argv, "dl:")) != -1 ) {
		switch (c) {
			case 'l':
				seg_length = atoi(optarg);
				break;
			case 'd':
				flat_dir_flag = 0;
				break;
			case '?':
			default:
				fprintf(STDERR, "main() : Unrecognised option : -%c\n", optopt);
				exit(1);
		}
	}
	
	/* The last arg is the filename */
	filename = argv[argc -1];
	fprintf(STDOUT, "INPUT FILE: %s\n", filename);

	if (read_pdb_file(filename) == FAIL) exit(1);
	if (0) fprintf(STDOUT, "main():    ... pdb file has been read\n");
	if (0) print_my_sequences();
	if (0) print_my_protein(&my_protein);
	
	create_files_for_each_segment(filename, seg_length);

	if (my_protein.model != NULL) free(my_protein.model);
	if (my_protein.atom != NULL) free(my_protein.atom);
	if (my_protein.sequence != NULL) free(my_protein.sequence);
	exit(0);
}
