/*
 *  ps_segment_files.c
 *  ProteinSplitter
 *
 *  Created by Naoto Morikawa on 7/22/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2007-12-21 Fix bug in create_entfile() and create_entfile_flat_dir()
 *  2007-12-21 Fix bug in print_segment_data()
 *  2008-01-02 version 0.2 release
 *  2008-01-02 .pdb file is supported in create_entfile() create_entfile_flat_dir()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>	/* mkdir */
#include <math.h>
#include "ps_main.h"
#include "ps_segment_files.h"
#include "ps_print.h"

extern struct pe_protein my_protein;
extern int flat_dir_flag;

void int_to_reschr(int res_int, char * r_chr)
{
	switch (res_int) {
		case (L_ALA):
			*r_chr = 'A';
			return;
		case (L_ASP):
			*r_chr = 'D';
			return;
		case (L_ARG):
			*r_chr = 'R';
			return;
		case (L_ASN):
			*r_chr = 'N';
			return;
		case (L_ASX):
			*r_chr = 'B';
			return;
		case (L_CYS):
			*r_chr = 'C';
			return;
		case (L_GLY):
			*r_chr = 'G';
			return;
		case (L_GLU):
			*r_chr = 'E';
			return;
		case (L_GLN):
			*r_chr = 'Q';
			return;
		case (L_GLX):
			*r_chr = 'Z';
			return;
		case (L_HIS):
			*r_chr = 'H';
			return;
		case (L_ILE):
			*r_chr = 'I';
			return;
		case (L_LEU):
			*r_chr = 'L';
			return;
		case (L_LYS):
			*r_chr = 'K';
			return;
		case (L_MET):
			*r_chr = 'M';
			return;
		case (L_PHE):
			*r_chr = 'F';
			return;
		case (L_PRO):
			*r_chr = 'P';
			return;
		case (L_SER):
			*r_chr = 'S';
			return;
		case (L_THR):
			*r_chr = 'T';
			return;
		case (L_TYR):
			*r_chr = 'Y';
			return;
		case (L_TRP):
			*r_chr = 'W';
			return;
		case (L_UNK):
			*r_chr = 'X';
			return;
		case (L_VAL):
			*r_chr = 'V';
			return;
		case (L_HYP):
			*r_chr = '0';
			return;
		default:
			fprintf(STDERR, "int_to_reschr() : ERROR invalid res_int = *%d*\n", res_int);
			exit(1);
	}
}

void int_to_resname(int res_int, char *r_name)
{
	switch (res_int) {
		case (L_ALA):strncpy(r_name, "ALA\0", 4);
		return;
	case (L_ASP):
		strncpy(r_name, "ASP\0", 4);
		return;
	case (L_ARG):
		strncpy(r_name, "ARG\0", 4);
		return;
	case (L_ASN):
		strncpy(r_name, "ASN\0", 4);
		return;
	case (L_ASX):
		strncpy(r_name, "ASX\0", 4);
		return;
	case (L_CYS):
		strncpy(r_name, "CYS\0", 4);
		return;
	case (L_GLY):
		strncpy(r_name, "GLY\0", 4);
		return;
	case (L_GLU):
		strncpy(r_name, "GLU\0", 4);
		return;
	case (L_GLN):
		strncpy(r_name, "GLN\0", 4);
		return;
	case (L_GLX):
		strncpy(r_name, "GLX\0", 4);
		return;
	case (L_HIS):
		strncpy(r_name, "HIS\0", 4);
		return;
	case (L_ILE):
		strncpy(r_name, "ILE\0", 4);
		return;
	case (L_LEU):
		strncpy(r_name, "LEU\0", 4);
		return;
	case (L_LYS):
		strncpy(r_name, "LYS\0", 4);
		return;
	case (L_MET):
		strncpy(r_name, "MET\0", 4);
		return;
	case (L_PHE):
		strncpy(r_name, "PHE\0", 4);
		return;
	case (L_PRO):
		strncpy(r_name, "PRO\0", 4);
		return;
	case (L_SER):
		strncpy(r_name, "SER\0", 4);
		return;
	case (L_THR):
		strncpy(r_name, "THR\0", 4);
		return;
	case (L_TYR):
		strncpy(r_name, "TYR\0", 4);
		return;
	case (L_TRP):
		strncpy(r_name, "TRP\0", 4);
		return;
	case (L_UNK):
		strncpy(r_name, "UNK\0", 4);
		return;
	case (L_VAL):
		strncpy(r_name, "VAL\0", 4);
		return;
	case (L_HYP):
		strncpy(r_name, "HYP\0", 4);
		return;
	default:
		fprintf(STDERR, "int_to_resname() : ERROR invalid res_int = *%d*\n", res_int);
		exit(1);
	}
}

int resname_to_int(char *r_name)
{
	switch (*r_name) {
		case ('A'):if (!strncmp("ALA", r_name, 3)) {
			return L_ALA;
		} else if (!strncmp("ASP", r_name, 3)) {
			return L_ASP;
		} else if (!strncmp("ARG", r_name, 3)) {
			return L_ARG;
		} else if (!strncmp("ASN", r_name, 3)) {
			return L_ASN;
		} else if (!strncmp("ASX", r_name, 3)) {
			return L_ASX;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('C'):
		if (!strncmp("CYS", r_name, 3)) {
			return L_CYS;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('G'):
		if (!strncmp("GLY", r_name, 3)) {
			return L_GLY;
		} else if (!strncmp("GLU", r_name, 3)) {
			return L_GLU;
		} else if (!strncmp("GLN", r_name, 3)) {
			return L_GLN;
		} else if (!strncmp("GLX", r_name, 3)) {
			return L_GLX;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('H'):
		if (!strncmp("HIS", r_name, 3)) {
			return L_HIS;
		} else if (!strncmp("HYP", r_name, 3)) {
			return L_HYP;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('I'):
		if (!strncmp("ILE", r_name, 3)) {
			return L_ILE;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('L'):
		if (!strncmp("LEU", r_name, 3)) {
			return L_LEU;
		} else if (!strncmp("LYS", r_name, 3)) {
			return L_LYS;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('M'):
		if (!strncmp("MET", r_name, 3)) {
			return L_MET;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('P'):
		if (!strncmp("PHE", r_name, 3)) {
			return L_PHE;
		} else if (!strncmp("PRO", r_name, 3)) {
			return L_PRO;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('S'):
		if (!strncmp("SER", r_name, 3)) {
			return L_SER;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('T'):
		if (!strncmp("THR", r_name, 3)) {
			return L_THR;
		} else if (!strncmp("TYR", r_name, 3)) {
			return L_TYR;
		} else if (!strncmp("TRP", r_name, 3)) {
			return L_TRP;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('U'):
		if (!strncmp("UNK", r_name, 3)) {
			return L_UNK;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	case ('V'):
		if (!strncmp("VAL", r_name, 3)) {
			return L_VAL;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
			exit(1);
		}

	default:
		fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%s*\n", r_name);
		exit(1);

	}
}

int search_segment(int init_nn_ct, int seq_ct, int seg_length, char * dir_name)
{
	int length_ct;
	int atm_ct;
	int rname_int;
	int i;

	length_ct = 0;
	for (atm_ct = init_nn_ct;
	     atm_ct <= my_protein.sequence[seq_ct].atom_tail - 2; atm_ct+=3) {
		if (my_protein.atom[atm_ct].atom_name != PE_NN) {
			dir_name[0] = '\0';
			return FAIL;
		}
		if (my_protein.atom[atm_ct + 1].atom_name != PE_CA) {
			dir_name[0] = '\0';
			return FAIL;
		}
		if (my_protein.atom[atm_ct + 2].atom_name != PE_CC) {
			dir_name[0] = '\0';
			return FAIL;
		}
		/* my_protein.atom[atm_ct].res_name */
		rname_int = resname_to_int(my_protein.atom[atm_ct].res_name);
		int_to_reschr(rname_int, dir_name + length_ct);
		length_ct++;
		if (length_ct >= seg_length) {
			dir_name[length_ct] = '\0';
			if (0) fprintf(STDERR, "search_segment(): DEBUG   dir_name = %s\n",
				dir_name);
			for (i = length_ct; i >= 0; i--)
				dir_name[i + 5] = dir_name[i];
			dir_name[0] = dir_name[5];
			dir_name[1] = '/';
			dir_name[2] = dir_name[5];
			dir_name[3] = dir_name[6];
			dir_name[4] = '/';
			if (0) fprintf(STDERR, "search_segment(): DEBUG   dir_name = %s\n",
				dir_name);
			return SUCCESS;
		}
	}
	dir_name[0] = '\0';
	return FAIL;
}

/* PDB data of a segment */
void print_segment_data(char * filename, FILE * fp_out, int init_nn_ct, int seg_length)
{
	FILE * fp_pdb;
	char PDBLine[LINE_WIDTH + 1];
	int atm_ct;
	
	/* HEADER */
	/* COMPND */	
	if ((fp_pdb = fopen(filename, "r")) == NULL) {
		fprintf(STDERR, "print_segment_data(): Can't open %s\n",
			filename);
		exit(1);
	}
	while (fgets(PDBLine, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (*PDBLine == 'C') {
			if (!strncmp("COMP", PDBLine, 4))
				fprintf(fp_out, "%s\n", PDBLine);
		} else if (*PDBLine == 'H') {
			if (!strncmp("HEAD", PDBLine, 4))
				fprintf(fp_out, "%s\n", PDBLine);
		} else if (*PDBLine == '\n') {
			continue;
		} else {
			break;
		}
	}
	if (fclose(fp_pdb)) {
		fprintf(STDERR, "print_segment_data(): Can't close %s\n",
			filename);
		abort();
	}

	/* MODEL */
	fprintf(fp_out,
		"MODEL        1                                                                  \n");
	/* ATOM */
	for (atm_ct = init_nn_ct;
	     atm_ct < init_nn_ct + 3.0 * seg_length; atm_ct++) {
		fprintf(fp_out,
			"ATOM  %5d ",
			my_protein.atom[atm_ct].serial_num);
		if (my_protein.atom[atm_ct].atom_name == PE_CC)
			fprintf(fp_out, " C  ");
		else if (my_protein.atom[atm_ct].atom_name == PE_CA)
			fprintf(fp_out, " CA ");
		else if (my_protein.atom[atm_ct].atom_name == PE_NN)
			fprintf(fp_out, " N  ");
		else if (my_protein.atom[atm_ct].atom_name == PE_RNA)
			fprintf(fp_out, " RNA");
		else
			fprintf(fp_out, " UN ");
		fprintf(fp_out,
			"%c%c%c%c %c%4d    %8.3f%8.3f%8.3f%6.2f                    \n",
			my_protein.atom[atm_ct].alt_loc,
			my_protein.atom[atm_ct].res_name[0],	// bug fix 20071221
			my_protein.atom[atm_ct].res_name[1],	// bug fix 20071221
			my_protein.atom[atm_ct].res_name[2],	// bug fix 20071221
			my_protein.atom[atm_ct].chain_id,	// bug fix 20071221
			my_protein.atom[atm_ct].res_seq,
			my_protein.atom[atm_ct].x,
			my_protein.atom[atm_ct].y,
			my_protein.atom[atm_ct].z,
			my_protein.atom[atm_ct].occupancy);
	}
	
	/* TER */
	fprintf(fp_out,
		"TER                                                                             \n");
	/* ENDMDL */
	fprintf(fp_out,
		"ENDMDL                                                                          \n");
	/* MASTER */
	/* END */
	fprintf(fp_out,
		"END                                                                             \n");
	return;
}

int create_entfile(char * fname, int init_nn_ct, int seq_ct, int seg_length,
		   char * dir_name)
{
	FILE * fp_out;
	char * outfile;
	char ext_name[20];
	char * filename;
	char * file_ext;
	int dirname_length, filename_length, ext_length;
	
	if (seg_length < 10) {
		dir_name[6] = '\0';
		if (mkdir(dir_name, 0755) != 0) {	/* seg5/A */
			if (0) fprintf(STDERR, 
					   "create_entfile() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
					   dir_name);
		} 
		dir_name[6] = '/';
		dir_name[9] = '\0';
		if (mkdir(dir_name, 0755) != 0) {	/* seg5/A/AB */
			if (0) fprintf(STDERR, 
					   "create_entfile() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
					   dir_name);
		}
		dir_name[9] = '/';
	} else if (seg_length < 100) {
		dir_name[7] = '\0';
		if (mkdir(dir_name, 0755) != 0) {	/* seg11/A */
			if (0) fprintf(STDERR,
				       "create_entfile() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
				       dir_name);
		} 
		dir_name[7] = '/';
		dir_name[10] = '\0';
		if (mkdir(dir_name, 0755) != 0) {	/* seg11/A/AB */
			if (0) fprintf(STDERR,
				       "create_entfile() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
				       dir_name);
		} else {
			if (0) fprintf(STDERR,
				       "create_entfile() : DEBUG mkdir(%s, 0755) succeed\n",
				       dir_name);
		}
		dir_name[10] = '/';
	} else {
		fprintf(STDERR,
			"create_entfile() : ERROR mkdir(%s, 0755) failed\n",
			dir_name);
		exit(1);
	}
	
	if (mkdir(dir_name, 0755) != 0) {
		if (0) fprintf(STDERR, 
				   "create_entfile() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
				   dir_name);
	}
	if (0) fprintf(STDOUT,
			   "create_entfile() : DEBUG dir_name = %s\n",
			   dir_name);
	
	/* Open the out file */
	dirname_length = (int) strlen(dir_name) + 1;
	filename = strrchr(fname, '/');	// s = "/***_a*_c*.ent"
	if (filename == NULL) filename = fname;
	else filename++;
	filename_length = (int) strlen(filename);
	sprintf(ext_name, "_a%d_c%d.pdb\0",		// 20080102
		my_protein.atom[init_nn_ct].serial_num, seq_ct);
	ext_length = strlen(ext_name) - 4;
	outfile = (char *) malloc((dirname_length + filename_length + ext_length)
				  * sizeof(char));	// Bug fix by NM on 20071221
	if (outfile == NULL) {
		fprintf(STDERR,
			"create_entfile() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, dir_name);
	outfile[dirname_length - 1] = '/';
	strcpy(outfile + dirname_length, filename);
	file_ext = strrchr(outfile, '.');	/* s = ".ent" */
	if (file_ext == NULL) {	// 20080102
		fprintf(STDERR,
			"create_entfile(): %s is not a \".ent\" file nor \".pdb\" file\n", fname);
		exit(1);
	}
	if ((strcmp(file_ext, ".ent") != 0) && (strcmp(file_ext, ".pdb") != 0)) {	// 20080102
		fprintf(STDERR,
			"create_entfile(): %s is not a \".ent\" file nor \".pdb\" file\n", fname);
		exit(1);
	}
	strcpy(file_ext, ext_name);
	if (0) fprintf(STDOUT, "OUTPUT FILE: %s\n", outfile);
	if ((fp_out = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "create_entfile(): can't open %s\n", outfile);
		exit(1);
	}
		    
	print_segment_data(fname, fp_out, init_nn_ct, seg_length);

	if (fclose(fp_out)) {
		fprintf(STDERR, "create_entfile(): Can't close %s\n", outfile);
		abort();
	}
	if (outfile != NULL) free(outfile);
	return SUCCESS;
}

int create_entfile_flat_dir(char * fname, int init_nn_ct, int seq_ct, int seg_length,
			char * dir_name)
{
	FILE * fp_out;
	char * outfile;
	char ext_name[20];
	char * filename;
	char * file_ext;
	int dirname_length, filename_length, ext_length;
	
	if (seg_length < 10) {	/* seg5/A */
		dir_name[4] = '\0';
	} else if (seg_length < 100) {	/* seg11/A */
		dir_name[5] = '\0';
	} else {
		fprintf(STDERR,
			"create_entfile_flat_dir() : ERROR mkdir(%s, 0755) failed\n",
			dir_name);
		exit(1);
	}

	if (mkdir(dir_name, 0755) != 0) {
		if (0) fprintf(STDERR, 
		       "create_entfile_flat_dir() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
		       dir_name);
	}
	if (0) fprintf(STDOUT,
	       "create_entfile_flat_dir() : DEBUG dir_name = %s\n",
	       dir_name);

	/* Open the out file */
	dirname_length = (int) strlen(dir_name) + 1;
	filename = strrchr(fname, '/');	// s = "/***_a*_c*.ent"
	if (filename == NULL) filename = fname;
	else filename++;
	filename_length = (int) strlen(filename);
	sprintf(ext_name, "_a%d_c%d.pdb\0",		// 20080102
		my_protein.atom[init_nn_ct].serial_num, seq_ct);
	ext_length = strlen(ext_name) - 4;
	outfile = (char *) malloc((dirname_length + filename_length + ext_length)
				  * sizeof(char));		// Bug fix by NM on 20071221
	if (outfile == NULL) {
		fprintf(STDERR,
			"create_entfile_flat_dir() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, dir_name);
	outfile[dirname_length - 1] = '/';
	strcpy(outfile + dirname_length, filename);
	file_ext = strrchr(outfile, '.');	/* s = ".ent" */
	if (file_ext == NULL) {		// 20080102
		fprintf(STDERR, "create_entfile_flat_dir(): %s is not a \".ent\" file nor \".pdb\" file\n",
			fname);
		exit(1);
	}
	if ((strcmp(file_ext, ".ent") != 0) && (strcmp(file_ext, ".pdb") != 0)) {	// 20080102
		fprintf(STDERR, "create_entfile_flat_dir(): %s is not a \".ent\" file nor \".pdb\" file\n",
			fname);
		exit(1);
	}
	strcpy(file_ext, ext_name);
	if (0) fprintf(STDOUT, "OUTPUT FILE: %s\n", outfile);
	if ((fp_out = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "create_entfile_flat_dir(): can't open %s\n", outfile);
		exit(1);
	}
					      
	print_segment_data(fname, fp_out, init_nn_ct, seg_length);
					      
	if (fclose(fp_out)) {
		fprintf(STDERR, "create_entfile_flat_dir(): Can't close %s\n", outfile);
		abort();
	}
	if (outfile != NULL) free(outfile);
	return SUCCESS;
}

#define TOP_DIR_NAME_LENGTH 11	/* "seg11/A/AB\n" */

void create_files_for_each_segment(char * filename, int seg_length)
{
	int mod_ct, seq_ct, atm_ct;
	int nn_ct, last_nn;
	int name_length;
	char * dir_name;
	int file_created;
	
	if (seg_length > 99) {
		fprintf(STDERR, 
			"create_files_for_each_segment() : ERROR seg_length =%d > 99\n",
			seg_length);
		exit(1);
	} else {
		if (0) fprintf(STDOUT, 
			"create_files_for_each_segment() : DEBUG seg_length = %d\n",
			seg_length);
	}
	dir_name = (char *) malloc((seg_length + TOP_DIR_NAME_LENGTH + 1)
				   * sizeof(char));
	if (dir_name == NULL) {
		fprintf(STDERR, 
			"create_files_for_each_segment() : ERROR dir_name[%d] allocation failed\n",
			seg_length);
		exit(1);
	}
	sprintf(dir_name, "seg%d\0", seg_length);
	if (0) fprintf(STDOUT,
			   "create_files_for_each_segment() : DEBUG dir_name = %s\n",
			   dir_name);
	if (mkdir(dir_name, 0755) != 0) {
		if (0) fprintf(STDERR, 
				   "create_files_for_each_segment() : DEBUG mkdir(%s, 0755) failed (It may already exist)\n",
				   dir_name);
	}
	name_length = strlen(dir_name);
	strncpy(dir_name + name_length, "/\0", 2);
	name_length++;
	
	file_created = 0;
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			nn_ct = 0;
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (my_protein.atom[atm_ct].atom_name == PE_NN)
					nn_ct++;
			}
			if (0) fprintf(STDOUT,
				"create_files_for_each_segment() : DEBUG nn_ct = %d in sequence %d\n",
				nn_ct, seq_ct);
			if (nn_ct < seg_length) continue;
			
			last_nn = nn_ct - seg_length + 1;
			atm_ct = my_protein.sequence[seq_ct].atom_head;
			for (nn_ct = 0; nn_ct <= last_nn; atm_ct++) {
				if (my_protein.atom[atm_ct].atom_name != PE_NN)
					continue;
				nn_ct++;
				if (search_segment(atm_ct, seq_ct, seg_length,
						dir_name + name_length) == FAIL)
					continue;
				if (flat_dir_flag == 0) {
					/* Structured dir: seg5/A/AB/ */
					if (create_entfile(filename, atm_ct, seq_ct,
						seg_length, dir_name) == FAIL)
						exit(1);
				} else {
					/* Flat dir: seg5/ */
					if (create_entfile_flat_dir(filename,
						atm_ct, seq_ct, seg_length,
						dir_name) == FAIL)
						exit(1);
				}
				file_created++;
			}
		}
	}
	if (file_created > 0) {
		if (DEBUG) fprintf(STDOUT,
				   "create_files_for_each_segment() : %d fragment files created by %s\n",
				   file_created, filename);
	} else {
		if (DEBUG) fprintf(STDOUT,
				   "create_files_for_each_segment() : 0 fragment files created by %s\n",
				   filename);
	}
	return;
}
